"""Taken from https://github.com/ZungBang/baculafs/blob/master/baculafs/Base64.py under GPLv3"""


# http://www.bacula.org/git/cgit.cgi/bacula/tree/bacula/src/findlib/attribs.c

def decode_base64(base64):
    """Bacula specific implementation of a base64 decoder"""
    digits = \
        ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M',
         'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
         'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm',
         'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z',
         '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '+', '/']

    base64_map = dict(zip(digits, xrange(0, 64)))

    value = 0
    first = 0
    neg = False

    if base64[0] == '-':
        neg = True
        first = 1

    for i in xrange(first, len(base64)):
        value = value << 6
        value += base64_map[base64[i]]

    return -value if neg else value



import string

from sqlalchemy.engine.reflection import Inspector


class LowerCaseInspector(Inspector):
    """Implements reflection inspector that reflects everything lowercase"""

#    def get_table_names(self, *a, **kw):
#        tables = super(LowerCaseInspector, self).get_table_names(*a, **kw)
#        return map(string.lower, tables)

    def get_columns(self, *a, **kw):
        columns = super(LowerCaseInspector, self).get_columns(*a, **kw)

        def lower_case(column):
            column['name'] = column['name'].lower()
            return column

        return map(lower_case, columns)

    def get_indexes(self, *a, **kw):
        columns = super(LowerCaseInspector, self).get_indexes(*a, **kw)

        def lower_case(column):
            column['name'] = column['name'].lower()
            column['column_names'] = map(string.lower, column['column_names'])
            return column

        return map(lower_case, columns)

    def get_foreign_keys(self, *a, **kw):
        columns = super(LowerCaseInspector, self).get_foreign_keys(*a, **kw)

        def lower_case(column):
            column['referred_table'] = column['referred_table'].lower()
            column['referred_columns'] = map(string.lower, column['referred_columns'])
            column['constrained_columns'] = map(string.lower, column['constrained_columns'])
            return column

        # ugly bacula bugix since sqlite 'LocationLog' foreignkey that depends on non-existent table LocationId
        columns = filter(lambda x: x['referred_table'] != u'LocationId', columns)

        return map(lower_case, columns)

    def get_pk_constraint(self, *a, **kw):
        columns = super(LowerCaseInspector, self).get_pk_constraint(*a, **kw)
        columns['constrained_columns'] = map(string.lower, columns['constrained_columns'])
        return columns



import unittest2

from mock import patch, MagicMock

from almir.scripts.configure_deploy import main


class TestConfigureDeploy(unittest2.TestCase):

    @patch('almir.scripts.configure_deploy.open', create=True)
    @patch('almir.lib.bconsole.BConsole.is_running')
    @patch('getpass.getpass')
    @patch('almir.scripts.configure_deploy.validate_engine')
    @patch('sys.stdout', spec=file)
    @patch('sys.stdin')
    def test_minimal_run(self, mock_stdin, mock_stdout, mock_validate_engine, mock_getpass, mock_is_running, mock_open):
        mock_stdin.readline.side_effect = ['\n', '\n', 'sqlite:///\n', '\n', '\n', '\n', '\n', '123123\n']
        mock_getpass.return_value = '123123'
        mock_is_running.return_value = True
        mock_open.return_value = MagicMock(spec=file)

        main()

        file_handle = mock_open.return_value.__enter__.return_value
        actual = file_handle.write.call_args[0][0]
        expected = """[buildout]
extends = buildout.d/production.cfg

[almir]
# define sql database connection as specified in
# http://docs.sqlalchemy.org/en/latest/core/engines.html#database-urls
#sqla_engine_url = sqlite:///%(here)s/bacula.db
#sqla_engine_url = postgresql+pg8000://bacula:<password>@hostname/<database>
#sqla_engine_url = mysql+mysqlconnector://bacula:<password>@hostname/<database>
sqla_engine_url = sqlite:///

# your timezone from TZ* column in table https://en.wikipedia.org/wiki/List_of_tz_database_time_zones
# if not set, defaults to system timezone
#timezone = Europe/Ljubljana
timezone =

# configure http daemon
host = 127.0.0.1
port = 2500

# bconsole director parameters
director_name = localhost-dir
director_port = 9101
director_address = localhost
director_password = 123123
        """
        self.assertEqual(actual.replace(' ', '').replace('\n', ''),
                         expected.replace(' ', '').replace('\n', ''),
                         "%s != %s" % (actual, expected))

    @patch('almir.scripts.configure_deploy.open', create=True)
    @patch('almir.lib.bconsole.BConsole.is_running')
    @patch('getpass.getpass')
    @patch('sqlalchemy.create_engine')  # TODO: rather patch engine
    @patch('sys.stdout', spec=file)
    @patch('sys.stdin')
    def test_custom_run(self, mock_stdin, mock_stdout, mock_create_engine, mock_getpass, mock_is_running, mock_open):
        mock_stdin.readline.side_effect = [
            'localhost\n',  # host
            'asd\n',  # port
            '99999\n',  # port
            '80\n',  # port
            '45001\n',  # port
            '\n',  # sqla
            'sqlite:///bacula.db\n',  # sqla
            'DST\n',  # timezone
            'Europe/Ljubljana\n',  # timezone
            'almir\n',  # dir name
            '8.8.8.8\n',  # dir address
            '8080\n',  # dir port
            '666\n',  # dir password
        ]
        engine = mock_create_engine.return_value
        engine.table_names.return_value = ['Client', 'Job', 'Media']
        mock_getpass.return_value = '666'
        mock_is_running.return_value = True
        mock_open.return_value = MagicMock(spec=file)

        main()

        file_handle = mock_open.return_value.__enter__.return_value
        actual = file_handle.write.call_args[0][0]
        expected = """
[buildout]
extends = buildout.d/production.cfg

[almir]
# define sql database connection as specified in
# http://docs.sqlalchemy.org/en/latest/core/engines.html#database-urls
#sqla_engine_url = sqlite:///%(here)s/bacula.db
#sqla_engine_url = postgresql+pg8000://bacula:<password>@hostname/<database>
#sqla_engine_url = mysql+mysqlconnector://bacula:<password>@hostname/<database>
sqla_engine_url = sqlite:///bacula.db

# your timezone from TZ* column in table https://en.wikipedia.org/wiki/List_of_tz_database_time_zones
# if not set, defaults to system timezone
#timezone = Europe/Ljubljana
timezone = Europe/Ljubljana

# configure http daemon
host = localhost
port = 45001

# bconsole director parameters
director_name = almir
director_port = 8080
director_address = 8.8.8.8
director_password = 666
        """
        self.assertEqual(actual.replace(' ', '').replace('\n', ''),
                         expected.replace(' ', '').replace('\n', ''),
                         "%s != %s" % (actual, expected))



import logging
import collections

from deform import Form, ValidationFailure
from pyramid.interfaces import IRoutesMapper
from pyramid.settings import asbool
from sqlalchemy import String, LargeBinary
from sqlalchemy.sql.expression import desc, or_
from sqlalchemy.sql.functions import sum, count
from sqlalchemy.orm import joinedload
from sqlalchemy.orm.attributes import InstrumentedAttribute

from almir.meta import DBSession, get_database_size
from almir.models import Job, Client, Log, Media, Storage, Pool, Status, File, Path, Filename
from almir.forms import JobSchema, MediaSchema, LogSchema
from almir.lib.console_commands import CONSOLE_COMMANDS
from almir.lib.bconsole import BConsole, DirectorNotRunning
from almir.lib.utils import render_rst_section, get_jinja_macro


log = logging.getLogger(__name__)


def dashboard(request):
    dbsession = DBSession()

    jobs = Job.get_last()
    running_jobs = Job.get_running()
    upcoming_jobs = Job.get_upcoming()
    try:
        director_version = '<span title="Connected to" class="label label-success">%s</span>' % BConsole().get_version()
    except DirectorNotRunning:
        director_version = '<span class="label label-important">Director not running!</span>'

    # statistics
    num_clients = dbsession.query(count(Client.clientid)).scalar()
    num_jobs = dbsession.query(count(Job.jobid)).scalar()
    num_volumes = dbsession.query(count(Media.mediaid)).scalar()
    sum_volumes = Media.format_byte_size(dbsession.query(sum(Media.volbytes)).scalar() or 0)
    database_size = get_database_size(DBSession.bind)

    return locals()


def about(request):
    changelog = render_rst_section('changelog.rst')
    about = render_rst_section('about.rst')
    return locals()


def console(request):
    command_array = ','.join(['"%s"' % name for name in CONSOLE_COMMANDS.keys()])
    console_commands = CONSOLE_COMMANDS
    return locals()


# restful generic view
class MixinView(object):
    model = None
    schema = None

    def __init__(self, request):
        self.request = request
        self.context = {
                        'appstruct': {},
        }

    def get_form(self):
        """Deals everything regarding forms for a request."""
        if not self.schema:
            return

        schema = self.schema()

        form = Form(schema.bind(**self.context),
                    buttons=[],
                    bootstrap_form_style='form-vertical')

        if self.request.query_string:
            controls = self.request.GET.items()

            try:
                self.context['appstruct'] = form.validate(controls)
                self.request.session[self.schema.__name__] = self.context['appstruct']
                self.request.session.save()
            except ValidationFailure, e:
                return e

        appstruct = self.request.session.get(self.schema.__name__, None)
        if appstruct:
            return form.render(appstruct)
        else:
            return form.render()

    def list(self):
        self.context['form'] = self.get_form()
        self.context['objects'] = self.model.get_list(appstruct=self.context['appstruct'])
        return self.context

    def detail(self):
        id_ = self.request.matchdict['id']
        self.context['object'] = self.model.get_one(id_)
        return self.context


class JobView(MixinView):
    model = Job
    schema = JobSchema

    def list(self):
        # TODO: cache this
        self.context['status_values'] = [('', '---')] + Status.get_values()
        return super(JobView, self).list()

    def detail(self):
        id_ = self.request.matchdict['id']
        self.context['files'] = File.query.filter(File.jobid == int(id_))\
                                          .options(joinedload('path'), joinedload('filename'))\
                                          .join('path').join('filename')
        return super(JobView, self).detail()

class ClientView(MixinView):
    model = Client

    def detail(self):
        # TODO: get jobs as join/subquery on client
        super(ClientView, self).detail()
        id_ = self.request.matchdict['id']
        jobs = Job.query.filter(Job.clientid == int(id_))
        self.context['jobs'] = jobs.options(joinedload(Job.status))\
                                   .order_by(desc(Job.schedtime))\
                                   .limit(50)  # TODO: dynamic datatables
        self.context['job_statistics'] = jobs.with_entities(count().label('num_jobs'), sum(Job.jobbytes).label('total_size_backups')).first(),
        self.context['last_successful_job'] = jobs.filter(Job.jobstatus == 'T').order_by(desc(Job.starttime)).first(),
        return self.context


class StorageView(MixinView):
    model = Storage


class VolumeView(MixinView):
    model = Media
    schema = MediaSchema

    def list(self):
        # UPSTREAM: we need to convert id to string, look at https://github.com/Pylons/deform/issues/81
        # TODO: cache
        self.context['storage_values'] = [('', '---')] + map(lambda x: (str(x[0]), x[1]), Storage.get_values())
        self.context['pool_values'] = [('', '---')] + map(lambda x: (str(x[0]), x[1]), Pool.get_values())
        return super(VolumeView, self).list()


class PoolView(MixinView):
    model = Pool


class LogView(MixinView):
    model = Log
    schema = LogSchema


bconsole_session = None
command_cache = collections.deque(maxlen=10)


def ajax_console_input(request):
    global bconsole_session
    # TODO: if status for client is requested, it might timeout after few seconds, but not 0.5
    # TODO: thread locking
    # TODO: implement session based on cookie
    # TODO: stderr?

    if not request.POST['bconsole_command'] and bconsole_session is not None:
        return {"commands": list(command_cache)}

    b = BConsole()
    bconsole_session, response = b.send_command_by_polling(request.POST['bconsole_command'], bconsole_session)

    if 'error' in response:
        command_cache.clear()

    if 'commands' in response:
        command_cache.extend(response['commands'])
    return response


def httpexception(context, request):
    request.response.status_int = context.code
    return {'context': context}


def datatables(request):
    """Implements server side interface for datatables.

    For field reference look at http://www.datatables.net/usage/server-side

    """
    link = get_jinja_macro('link')

    q = request.GET
    iDisplayStart = int(q.get('iDisplayStart'))
    iDisplayLength = int(q.get('iDisplayLength'))
    iColumns = int(q.get('iColumns'))
    sColumns = q.get('sColumns').split(',')
    sSearch = q.get('sSearch')
    #bRegex = q.get('bRegex')
    #iSortingCols = int(q.get('iSortingCols'))
    sEcho = int(q.get('sEcho'))

    # get view from referrer url
    for route in request.registry.getUtility(IRoutesMapper).routelist:
        match = route.match(q.get('referrer'))
        if match is not None:
            break

    for intro in request.registry.introspector.get_category('views'):
        if intro['introspectable']['route_name'] == route.name:
            break

    view = intro['introspectable']['callable'](request)
    # TODO: views will do form filtering, meaning total_records will give false information
    if intro['introspectable']['attr'] == 'list':
        query = view.list()[q.get('context', 'objects')]
    else:
        request.matchdict = match
        query = view.detail()[q.get('context', 'object')]

    # get model from query entity
    model = query._entities[0].mapper.class_

    total_records = query.count()
    search_filter = []

    for i in range(iColumns):
        mDataProp = q.get('mDataProp_%d' % i)
        try:
            field = sColumns[i] or mDataProp
        except IndexError:
            # no information about columns
            field = mDataProp

        # handle search
        bSearchable = asbool(q.get('bSearchable_%d' % i))
        if bSearchable:
            # TODO: handle regex
            # TODO: column specific search: col_sSearch = q.get('sSearch_%d' % i)
            # TODO refactor to models to provide generic search and edge cases for specific models
            if sSearch:
                if field == 'path.path;filename.name':
                    search_filter.append((Path.path + Filename.name).ilike('%' + sSearch + '%'))
                else:
                    column = getattr(model, field)
                    if isinstance(column.property.columns[0].type, (String, LargeBinary)):
                        search_filter.append(column.ilike('%' + sSearch + '%'))
                    # TODO: support more type searches

        # handle sorting
        iSortCol = q.get('iSortCol_%d' % i)
        bSortable = asbool(q.get('bSortable_%d' % i))
        if iSortCol and bSortable:
            # TODO refactor to models to provide generic ordering and edge cases for specific models
            for prop in field.split(';'):
                # get colum for sorting
                order_column = model
                for attr in prop.split('.'):
                    if isinstance(order_column, InstrumentedAttribute):
                        # handle relations
                        order_column = order_column.property.mapper.class_
                    order_column = getattr(order_column, attr)

                # sort
                if q.get('sSortDir_%d' % i) == 'desc':
                    query = query.order_by(desc(order_column))
                else:
                    query = query.order_by(order_column)

    query = query.filter(or_(*search_filter))
    total_display_records = query.count()

    aaData = []

    for result in query[iDisplayStart:iDisplayStart + iDisplayLength]:
        obj = {}
        for i in range(iColumns):
            mDataProp = q.get('mDataProp_%d' % i)
            options = getattr(result, 'render_%s' % mDataProp)(request) or {}
            obj[mDataProp] = link(options)
            if options.get('cssclass'):
                obj['DT_RowClass'] = options['cssclass']
        aaData.append(obj)

    return {
            'sEcho': sEcho,
            'iTotalDisplayRecords': total_display_records,
            'iTotalRecords': total_records,
            'aaData': aaData,
    }



import os
import random
import hashlib
import pkg_resources

from pyramid_beaker import BeakerSessionFactoryConfig
from pyramid.config import Configurator
from pyramid.events import BeforeRender
from pyramid.httpexceptions import HTTPError
from sqlalchemy.dialects import sqlite
from sqlalchemy.types import INTEGER

from almir.meta import initialize_sql


try:
    __version__ = pkg_resources.get_distribution("almir").version
except:
    __version__ = ''


def navigation_tree(event):
    """Generate navigation data"""
    request = event['request']
    if not request:
        # debugger does not provide request
        return  # PRAGMA: no cover
    event['navigation_tree'] = [
        dict(name="Dashboard", url=request.route_url('dashboard')),
        dict(name="Clients", url=request.route_url('client_list')),
        # TODO: add later dict(name="FileSets", url=request.route_url('fileset_list')),
        dict(name="Jobs", url=request.route_url('job_list')),
        dict(name="Volumes", url=request.route_url('volume_list')),
        dict(name="Pools", url=request.route_url('pool_list')),
        dict(name="Storages", url=request.route_url('storage_list')),
        dict(name="Logs", url=request.route_url('log')),
        dict(name="Console", url=request.route_url('console')),
    ]

    try:
        event['current_url'] = request.current_route_url()
    except ValueError:
        event['current_url'] = ''


def main(global_config, **settings):
    """ This function returns a Pyramid WSGI application. """
    # provide aliases for sqlite types
    # to silence warnings
    sqlite.base.ischema_names['TINYINT'] = INTEGER
    sqlite.base.ischema_names['BIGINT'] = INTEGER

    initialize_sql(settings)

    config = Configurator(settings=settings)
    config.include('pyramid_jinja2')
    config.include('pyramid_tm')
    config.include('pyramid_beaker')
    config.include('deform_bootstrap')

    # setup beaker
    beaker_dict = dict(
                   type="memory",
                   lock_dir=os.path.join(global_config.get('here', '.'), 'session_lock'),
                   secret=hashlib.sha1(str(random.getrandbits(100))).hexdigest(),
                   )
    config.set_session_factory(BeakerSessionFactoryConfig(**beaker_dict))

    # events
    config.add_subscriber(navigation_tree, BeforeRender)

    # static
    config.add_static_view('static', 'almir:static', cache_max_age=3600)
    config.add_static_view('static_deform', 'deform:static')

    # routes
    config.add_route('dashboard', '/')
    config.add_view('almir.views.dashboard',
                    route_name='dashboard',
                    renderer='templates/dashboard.jinja2')
    config.add_route('about', '/about/')
    config.add_view('almir.views.about',
                    route_name='about',
                    renderer='templates/about.jinja2')
    config.add_route('log', '/log/')
    config.add_view('almir.views.LogView',
                    attr="list",
                    route_name='log',
                    renderer='templates/log.jinja2')
    config.add_route('console', '/console/')
    config.add_view('almir.views.console',
                    route_name='console',
                    renderer='templates/console.jinja2')
    config.add_route('console_ajax', '/console/input/')
    config.add_view('almir.views.ajax_console_input',
                    route_name='console_ajax',
                    renderer='json',
                    request_method='POST')
    config.add_route('datatables', '/datatables/')
    config.add_view('almir.views.datatables',
                    route_name='datatables',
                    renderer='json',
                    request_method='GET')

    # exception handling views
    config.add_view('almir.views.httpexception',
                    context=HTTPError,
                    renderer='templates/httpexception.jinja2')
    config.add_notfound_view('almir.views.httpexception',
                             renderer='templates/httpexception.jinja2',
                             append_slash=True)

    # RESTful resources
    for name in ['job', 'client', 'storage', 'volume', 'pool']:
        for action, url in [('list', '/%s/' % name), ('detail', '/%s/{id:\d+}/' % name)]:
            config.add_route('%s_%s' % (name, action), url)
            config.add_view(
                'almir.views.%sView' % name.title(),
                route_name='%s_%s' % (name, action),
                attr=action,
                renderer='templates/%s_%s.jinja2' % (name, action),
            )

    return config.make_wsgi_app()



import os
import urlparse
import json
import logging.config
from subprocess import Popen, PIPE

from mock import patch
from paste.deploy.loadwsgi import appconfig
from webtest import TestApp

from almir import main
from almir.tests import AlmirTestCase, assert_num_of_queries


here = os.path.dirname(__file__)


class FunctionalTests(AlmirTestCase):
    """Runs tests under multiple datasets"""
    config_file = os.path.join(here, '../../', 'development.ini')
    settings = appconfig('config:' + config_file)
    db_conns = {
        'sqlite': 'sqlite:///%s' % os.path.join(here, 'fixtures', 'sqlite.db'),
        'mysql': 'mysql+mysqlconnector://root@localhost/bacula',
        'postgresql': 'postgresql+pg8000://postgres@localhost/bacula',
    }
    database = os.environ.get('DATABASE', 'sqlite')

    @classmethod
    def setUpClass(cls):
        super(FunctionalTests, cls).setUpClass()
        engine = os.environ.get('ENGINE', '')

        if engine:  # PRAGMA: no cover
            cls.settings['sqlalchemy.url'] = engine
            cls.database = urlparse.urlparse(engine).scheme.split('+')[0]
        else:
            cls.settings['sqlalchemy.url'] = cls.db_conns[cls.database]

        cls.app = main({}, **cls.settings)
        cls.testapp = TestApp(cls.app)
        logging.config.fileConfig(cls.config_file)

    @assert_num_of_queries(8, mysql=7, postgresql=7)
    @patch('almir.lib.bconsole.BConsole.start_process')
    @patch('almir.lib.bconsole.BConsole.get_version')
    def test_root(self, mock_get_version, mock_process):
        mock_process.return_value = Popen(['cat'], stdout=PIPE, stdin=PIPE, stderr=PIPE)

        res = self.testapp.get('/', status=200)
        self.failUnless('Last 5 finished Jobs' in res.body)
        self.failUnless('Running Jobs' in res.body)

    @assert_num_of_queries(0)
    def test_about(self):
        res = self.testapp.get('/about/', status=200)
        self.failUnless('Domen' in res.body)

    @assert_num_of_queries(0)
    def test_logs(self):
        res = self.testapp.get('/log/', status=200)
        self.failUnless('Log' in res.body)

    @assert_num_of_queries(0)
    def test_console(self):
        res = self.testapp.get('/console/', status=200)
        self.failUnless('Console' in res.body)

    @assert_num_of_queries(0)
    @patch('almir.lib.bconsole.BConsole.start_process')
    def test_console_ajax(self, mock_process):
        mock_process.return_value = Popen(['cat'], stdout=PIPE, stdin=PIPE, stderr=PIPE)

        res = self.testapp.post('/console/input/', {'bconsole_command': 'version'}, status=200)
        self.assertTrue('commands' in res.body)

        res = self.testapp.post('/console/input/', {'bconsole_command': ''}, status=200)
        self.assertTrue('commands' in res.body)

# restful

    @assert_num_of_queries(1)
    def test_restful_404(self):
        self.testapp.get('/client/9999/', status=404)

    @assert_num_of_queries(0)
    def test_append_slash(self):
        res = self.testapp.get('/about', status=302)
        res.follow(status=200)

    @assert_num_of_queries(1)
    def test_jobs(self):
        res = self.testapp.get('/job/', status=200)
        self.failUnless('Jobs' in res.body)

    @assert_num_of_queries(1)
    def test_job(self):
        ids = {
               'sqlite': 15,
        }
        id_ = ids.get(self.database, 1)
        res = self.testapp.get('/job/%d/' % id_, status=200)
        self.failUnless('Log' in res.body)

    @assert_num_of_queries(1)
    def test_clients(self):
        res = self.testapp.get('/client/', status=200)
        self.failUnless('Clients' in res.body)

    @assert_num_of_queries(4)
    def test_client(self):
        res = self.testapp.get('/client/1/', status=200)
        self.failUnless('Log' in res.body)

    @assert_num_of_queries(1)
    def test_storages(self):
        res = self.testapp.get('/storage/', status=200)
        self.failUnless('Storage' in res.body)

    @assert_num_of_queries(1)
    def test_storage(self):
        res = self.testapp.get('/storage/1/', status=200)
        self.failUnless('Storage' in res.body)

    @assert_num_of_queries(3)
    def test_volumes(self):
        res = self.testapp.get('/volume/', status=200)
        self.failUnless('Volumes' in res.body)

    @assert_num_of_queries(1)
    def test_volume(self):
        res = self.testapp.get('/volume/1/', status=200)
        self.failUnless('Volume' in res.body)

    @assert_num_of_queries(1)
    def test_pools(self):
        res = self.testapp.get('/pool/', status=200)
        self.failUnless('Pools' in res.body)

    @assert_num_of_queries(1)
    def test_pool(self):
        res = self.testapp.get('/pool/1/', status=200)
        self.failUnless('Pool' in res.body)

    @assert_num_of_queries(4)
    def test_datatables_jobs(self):
        params = {
            "sEcho": "1",
            "iColumns": "9",
            "sColumns": "jobid,status.joblongstatus,,,,client.clientid,,,",
            "iDisplayStart": "0",
            "iDisplayLength": "50",
            "mDataProp_0": "name",
            "mDataProp_1": "status",
            "mDataProp_2": "type",
            "mDataProp_3": "level",
            "mDataProp_4": "jobfiles",
            "mDataProp_5": "client_name",
            "mDataProp_6": "starttime",
            "mDataProp_7": "duration",
            "mDataProp_8": "joberrors",
            "sSearch": "",
            "bRegex": "false",
            "sSearch_0": "",
            "bRegex_0": "false",
            "bSearchable_0": "true",
            "sSearch_1": "",
            "bRegex_1": "false",
            "bSearchable_1": "true",
            "sSearch_2": "",
            "bRegex_2": "false",
            "bSearchable_2": "true",
            "sSearch_3": "",
            "bRegex_3": "false",
            "bSearchable_3": "true",
            "sSearch_4": "",
            "bRegex_4": "false",
            "bSearchable_4": "true",
            "sSearch_5": "",
            "bRegex_5": "false",
            "bSearchable_5": "true",
            "sSearch_6": "",
            "bRegex_6": "false",
            "bSearchable_6": "true",
            "sSearch_7": "",
            "bRegex_7": "false",
            "bSearchable_7": "false",
            "sSearch_8": "",
            "bRegex_8": "false",
            "bSearchable_8": "true",
            "iSortingCols": "1",
            "iSortCol_0": "6",
            "sSortDir_0": "desc",
            "bSortable_0": "true",
            "bSortable_1": "true",
            "bSortable_2": "true",
            "bSortable_3": "true",
            "bSortable_4": "true",
            "bSortable_5": "true",
            "bSortable_6": "true",
            "bSortable_7": "false",
            "bSortable_8": "true",
            "referrer": "/job/",
            "_charset_": "",
            "__formid__": "deform",
            "state": "finished+running",
            "type": "",
            "status": "",
        }
        res = self.testapp.get('/datatables/', params, status=200)
        self.assertTrue(res.json)

        d = json.loads(res.body)
        self.assertEqual(d['sEcho'], 1)
        self.assertEqual(d['iTotalDisplayRecords'], d['iTotalRecords'])
        self.assertEqual(d['iTotalDisplayRecords'], len(d['aaData']))

    @assert_num_of_queries(4)
    def test_datatables_jobs_form(self):
        params = {
            "sEcho": "1",
            "iColumns": "9",
            "sColumns": "jobid,status.joblongstatus,,,,client.clientid,,,",
            "iDisplayStart": "0",
            "iDisplayLength": "50",
            "mDataProp_0": "name",
            "mDataProp_1": "status",
            "mDataProp_2": "type",
            "mDataProp_3": "level",
            "mDataProp_4": "jobfiles",
            "mDataProp_5": "client_name",
            "mDataProp_6": "starttime",
            "mDataProp_7": "duration",
            "mDataProp_8": "joberrors",
            "sSearch": "",
            "bRegex": "false",
            "sSearch_0": "",
            "bRegex_0": "false",
            "bSearchable_0": "true",
            "sSearch_1": "",
            "bRegex_1": "false",
            "bSearchable_1": "true",
            "sSearch_2": "",
            "bRegex_2": "false",
            "bSearchable_2": "true",
            "sSearch_3": "",
            "bRegex_3": "false",
            "bSearchable_3": "true",
            "sSearch_4": "",
            "bRegex_4": "false",
            "bSearchable_4": "true",
            "sSearch_5": "",
            "bRegex_5": "false",
            "bSearchable_5": "true",
            "sSearch_6": "",
            "bRegex_6": "false",
            "bSearchable_6": "true",
            "sSearch_7": "",
            "bRegex_7": "false",
            "bSearchable_7": "false",
            "sSearch_8": "",
            "bRegex_8": "false",
            "bSearchable_8": "true",
            "iSortingCols": "1",
            "iSortCol_0": "6",
            "sSortDir_0": "desc",
            "bSortable_0": "true",
            "bSortable_1": "true",
            "bSortable_2": "true",
            "bSortable_3": "true",
            "bSortable_4": "true",
            "bSortable_5": "true",
            "bSortable_6": "true",
            "bSortable_7": "false",
            "bSortable_8": "true",
            "referrer": "/job/",
            "_charset_": "",
            "__formid__": "deform",
            "state": "finished+running",
            "type": "R",
            "status": "C",
        }
        res = self.testapp.get('/datatables/', params, status=200)
        self.assertTrue(res.json)

        d = json.loads(res.body)
        self.assertEqual(d['sEcho'], 1)
        self.assertEqual(d['iTotalRecords'], 0)
        self.assertEqual(len(d['aaData']), 0)

    @assert_num_of_queries(4)
    def test_datatables_files(self):
        ids = {
               'sqlite': 15,
        }
        id_ = ids.get(self.database, 1)
        params = {
            "sEcho": "1",
            "iColumns": "5",
            "sColumns": "path.path;filename.name,,,,",
            "iDisplayStart": "0",
            "iDisplayLength": "50",
            "mDataProp_0": "filename",
            "mDataProp_1": "size",
            "mDataProp_2": "mode",
            "mDataProp_3": "uid",
            "mDataProp_4": "gid",
            "sSearch": "",
            "bRegex": "false",
            "sSearch_0": "",
            "bRegex_0": "false",
            "bSearchable_0": "true",
            "sSearch_1": "",
            "bRegex_1": "false",
            "bSearchable_1": "false",
            "sSearch_2": "",
            "bRegex_2": "false",
            "bSearchable_2": "false",
            "sSearch_3": "",
            "bRegex_3": "false",
            "bSearchable_3": "false",
            "sSearch_4": "",
            "bRegex_4": "false",
            "bSearchable_4": "false",
            "iSortingCols": "1",
            "iSortCol_0": "0",
            "sSortDir_0": "asc",
            "bSortable_0": "true",
            "bSortable_1": "false",
            "bSortable_2": "false",
            "bSortable_3": "false",
            "bSortable_4": "false",
            "referrer": "/job/%d/" % id_,
            "context": "files",
        }
        res = self.testapp.get('/datatables/', params, status=200)
        self.assertTrue(res.json)

        d = json.loads(res.body)
        self.assertEqual(d['sEcho'], 1)
        self.assertEqual(d['iTotalDisplayRecords'], d['iTotalRecords'])
        self.assertEqual(d['iTotalDisplayRecords'], len(d['aaData']))

    @assert_num_of_queries(4)
    def test_datatables_files_search(self):
        ids = {
               'sqlite': 15,
        }
        id_ = ids.get(self.database, 1)
        params = {
            "sEcho": "1",
            "iColumns": "5",
            "sColumns": "path.path;filename.name,,,,",
            "iDisplayStart": "0",
            "iDisplayLength": "50",
            "mDataProp_0": "filename",
            "mDataProp_1": "size",
            "mDataProp_2": "mode",
            "mDataProp_3": "uid",
            "mDataProp_4": "gid",
            "sSearch": "blablabla",
            "bRegex": "false",
            "sSearch_0": "",
            "bRegex_0": "false",
            "bSearchable_0": "true",
            "sSearch_1": "",
            "bRegex_1": "false",
            "bSearchable_1": "false",
            "sSearch_2": "",
            "bRegex_2": "false",
            "bSearchable_2": "false",
            "sSearch_3": "",
            "bRegex_3": "false",
            "bSearchable_3": "false",
            "sSearch_4": "",
            "bRegex_4": "false",
            "bSearchable_4": "false",
            "iSortingCols": "1",
            "iSortCol_0": "0",
            "sSortDir_0": "asc",
            "bSortable_0": "true",
            "bSortable_1": "false",
            "bSortable_2": "false",
            "bSortable_3": "false",
            "bSortable_4": "false",
            "referrer": "/job/%d/" % id_,
            "context": "files",
        }
        res = self.testapp.get('/datatables/', params, status=200)
        self.assertTrue(res.json)

        d = json.loads(res.body)
        self.assertEqual(d['sEcho'], 1)
        self.assertEqual(d['iTotalDisplayRecords'], len(d['aaData']))
        self.assertEqual(len(d['aaData']), 0)

    @assert_num_of_queries(3)
    def test_datatables_logs(self):
        params = {
            "sEcho": "1",
            "iColumns": "3",
            "sColumns": "",
            "iDisplayStart": "0",
            "iDisplayLength": "1",
            "mDataProp_0": "jobid",
            "mDataProp_1": "time",
            "mDataProp_2": "logtext",
            "sSearch": "Found to prune",
            "bRegex": "false",
            "sSearch_0": "",
            "bRegex_0": "false",
            "bSearchable_0": "true",
            "sSearch_1": "",
            "bRegex_1": "false",
            "bSearchable_1": "false",
            "sSearch_2": "",
            "bRegex_2": "false",
            "bSearchable_2": "true",
            "iSortingCols": "1",
            "iSortCol_0": "1",
            "sSortDir_0": "desc",
            "bSortable_0": "true",
            "bSortable_1": "true",
            "bSortable_2": "true",
            "referrer": "/log/",
        }
        res = self.testapp.get('/datatables/', params, status=200)
        self.assertTrue(res.json)

        d = json.loads(res.body)
        self.assertEqual(d['sEcho'], 1)
        self.assertNotEqual(d['iTotalDisplayRecords'], d['iTotalRecords'])
        self.assertNotEqual(d['iTotalDisplayRecords'], len(d['aaData']))


if os.environ.get('ALL_DATABASES', ''):  # PRAGMA: no coverage
    MySQLFunctionalTests = type('MySQLFunctionalTests', (FunctionalTests,), {'database': 'mysql'})
    PGFunctionalTests = type('PGFunctionalTests', (FunctionalTests,), {'database': 'postgresql'})



import functools
import unittest2

from pyramid import testing
from sqlalchemy import event
from sqlalchemy.engine import Engine


def assert_num_of_queries(num_of_queries, **dbs):
    def dec(func):
        @functools.wraps(func)
        def wrapper(*a, **kw):
            self = a[0]
            num_of_q = dbs.get(self.database, num_of_queries)
            before = self.num_of_sql_queries
            func(*a, **kw)
            after = self.num_of_sql_queries
            self.assertEqual(after - before, num_of_q, "%s: Expected %d queries, got %d (on %s)" % (func.__name__, num_of_q, after - before, self.database))
        return wrapper
    return dec


class AlmirTestCase(unittest2.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.config = testing.setUp()

        # prepare query counting
        cls.num_of_sql_queries = 0
        event.listen(Engine, "after_cursor_execute", cls.count_query)

    @classmethod
    def tearDown(cls):
        testing.tearDown()

    @classmethod
    def count_query(cls, *a, **kw):
        cls.num_of_sql_queries += 1



import datetime

from almir.lib.utils import convert_timezone
from almir.tests.test_functional import AlmirTestCase


class TestUtils(AlmirTestCase):

    def test_timezone(self):
        self.assertEqual(convert_timezone(None), None)
        self.assertNotEqual(convert_timezone(datetime.datetime.now()).tzinfo, None)



# taken from https://bitbucket.org/zzzeek/sqlalchemy/src/3b458030a0f3/examples/declarative_reflection/declarative_reflection.py
from sqlalchemy import *
from sqlalchemy.orm import *
from sqlalchemy.orm.util import _is_mapped_class
from sqlalchemy.ext.declarative import declarative_base, declared_attr


class DeclarativeReflectedBase(object):
    _mapper_args = []

    @classmethod
    def __mapper_cls__(cls, *args, **kw):
        """Declarative will use this function in lieu of
        calling mapper() directly.

        Collect each series of arguments and invoke
        them when prepare() is called.
        """

        cls._mapper_args.append((args, kw))

    @classmethod
    def prepare(cls, engine):
        """Reflect all the tables and map !"""
        while cls._mapper_args:
            args, kw = cls._mapper_args.pop()
            klass = args[0]
            # autoload Table, which is already
            # present in the metadata.  This
            # will fill in db-loaded columns
            # into the existing Table object.
            if args[1] is not None:
                table = args[1]
                Table(table.name,
                    cls.metadata,
                    extend_existing=True,
                    autoload_replace=False,
                    autoload=True,
                    autoload_with=engine,
                    schema=table.schema)

            # see if we need 'inherits' in the
            # mapper args.  Declarative will have
            # skipped this since mappings weren't
            # available yet.
            for c in klass.__bases__:
                if _is_mapped_class(c):  # pragma: nocover
                    kw['inherits'] = c
                    break

            klass.__mapper__ = mapper(*args, **kw)



#import os.path
#import pickle
import logging
#import hashlib
#import sqlite3
from datetime import datetime

from pyramid.httpexceptions import exception_response
from sqlalchemy import engine_from_config
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import scoped_session
from sqlalchemy.orm import sessionmaker
from zope.sqlalchemy import ZopeTransactionExtension
from webhelpers.date import distance_of_time_in_words
from webhelpers.number import format_byte_size

from almir.lib.sqlalchemy_declarative_reflection import DeclarativeReflectedBase
from almir.lib.sqlalchemy_lowercase_inspector import LowerCaseInspector
from almir.lib.utils import timedelta_to_seconds, convert_timezone


log = logging.getLogger(__name__)

DBSession = scoped_session(sessionmaker(extension=ZopeTransactionExtension()))
Base = declarative_base(cls=DeclarativeReflectedBase)


class ModelMixin(object):
    query = DBSession.query_property()

    @declared_attr
    def __tablename__(cls):
        if cls.metadata.bind.dialect.name == 'postgresql':  # PRAGMA: no cover
            return cls.__name__.lower()
        else:
            return cls.__name__

    @classmethod
    def get_list(cls, **kw):
        return cls.query

    @classmethod
    def get_one(cls, id_=None, query=None):
        if id_ is not None:
            query = cls.query.get(int(id_))

        if query == None:
            raise exception_response(404)
        else:
            return query

    @staticmethod
    def format_byte_size(size):
        if size:
            # we use float since postgres driver will return decimal
            return format_byte_size(float(size))
        else:
            return format_byte_size(0)

    @staticmethod
    def render_distance_of_time_in_words(dt_from, dt_to=None):
        if not dt_from:
            return

        if dt_from.tzinfo is None:
            dt_from = convert_timezone(dt_from)

        if dt_to is None:
            return {'text': distance_of_time_in_words(dt_from, convert_timezone(datetime.now())) + ' ago', 'data_numeric': dt_from.strftime('%s')}
        else:
            if dt_to.tzinfo is None:
                dt_to = convert_timezone(dt_to)
            return {'text': distance_of_time_in_words(dt_from, dt_to),
                    'data_numeric': -timedelta_to_seconds(dt_to - dt_from)}


# TODO: cache for 5min
def get_database_size(engine):
    """Returns human formatted SQL database size.

    Example: 3.04 GB
    """
    if engine.name == 'sqlite':
        size_bytes = engine.execute('PRAGMA page_count;').scalar() * engine.execute('PRAGMA page_size;').scalar()
    elif engine.name == 'mysql':
        size_bytes = engine.execute('SELECT sum(ROUND((DATA_LENGTH + INDEX_LENGTH),2)) AS Size '
                                    'FROM INFORMATION_SCHEMA.TABLES where TABLE_SCHEMA like "%s";' % engine.url.database).scalar()
    elif engine.name == 'postgresql':
        size_bytes = engine.execute('SELECT pg_database_size(\'%s\');' % engine.url.database).scalar()
    else:
        raise NotImplemented
    return format_byte_size(float(size_bytes))


def initialize_sql(settings):
    # make sure metadata is populated

    engine = engine_from_config(settings, prefix='sqlalchemy.', encoding='utf-8')

    # monkey patch inspector to reflect lowercase tables/columns since sqlite/mysql have mixed case
    # while postgres has lowercase tables/columns
    engine.dialect.inspector = LowerCaseInspector

    DBSession.configure(bind=engine)
    DBSession.bind = engine
    Base.metadata.bind = engine

    import almir.models

    # cache (pickle) metadata
    Base.prepare(engine)

    # hash engine paramters so we don't cache wrong metadata
    #engine_hash = hashlib.md5(str(engine.url)).hexdigest()
    # TODO: figure out a way for sqlalchemy_declarative_reflection to work with metadata caching
    # TODO: configure with .ini
    #cachefile = os.path.join(os.path.dirname(__name__), 'db.metada.cache.%s' % engine_hash)
    #if os.path.isfile(cachefile):
    #    log.info('Loading database schema from cache file: %s', cachefile)
    #with open(cachefile, 'r') as cache:
    #Base.metadata = pickle.load(cache)
    #else:
    #log.info('Generating database schema cache file: %s', cachefile)
    #    with open(cachefile, 'w') as cache:
    #        pickle.dump(Base.metadata, cache)



"""Models generated from bacula-dir-postgresql 5.0.2"""
import datetime
import stat

from jinja2 import Markup
from sqlalchemy import Column
from sqlalchemy.orm import relationship, joinedload, joinedload_all
from sqlalchemy.sql.expression import desc
from sqlalchemy.sql import functions as func
from webhelpers.date import distance_of_time_in_words

from almir.meta import Base, ModelMixin, DBSession
from almir.lib.bacula_base64 import decode_base64
from almir.lib.bconsole import BConsole
from almir.lib.sqlalchemy_custom_types import BaculaDateTime
from almir.lib.utils import nl2br, yesno


# defined in http://www.bacula.org/git/cgit.cgi/bacula/tree/bacula/src/jcr.h
LEVELS = {
    'F': 'Full',
    'I': 'Incremental',  # since last backup
    'D': 'Differential',  # since last full backup
    'S': 'Since',
    'C': 'Verify from catalog',
    'V': 'Verify safe from DB',
    'O': 'Verify Volume to catalog',
    'd': 'Verify Disk to catalog',
    'A': 'Verify data on volume',
    'B': 'Base level job',
    ' ': 'Restore/Admin',
    'f': 'Virtual full',
}
TYPES = {
    'B': 'Backup',
    'M': 'Migrated',
    'V': 'Verify',
    'R': 'Restore',
    'U': 'Console',
    'D': 'Admin',
    'A': 'Archive',
    'C': 'Copy of a Job',
    'c': 'Copy',
    'g': 'Migration',
    'S': 'Scan',
}
# TODO: parse from bacula/bacula/src/lib/util.c
VOLUME_STATUS_SEVERITY = {
    "Append": 'ok',
    "Archive": 'unknown',
    "Disabled": 'unknown',
    "Full": 'unknown',
    "Used": 'unknown',
    "Cleaning": 'unknown',
    "Purged": 'unknown',
    "Recycle": 'unknown',
    "Read-Only": 'unknown',
    "Error": 'error',
}


class Status(ModelMixin, Base):
    """
        Column     |     Type     | Modifiers
    ---------------+--------------+-----------
     jobstatus     | character(1) | not null
     jobstatuslong | text         |
     severity      | integer      |

    Indexes:
        "status_pkey" PRIMARY KEY, btree (jobstatus)

    """

    @classmethod
    def get_values(cls):
        return cls.query.with_entities(cls.jobstatus, cls.jobstatuslong).all()


class Client(ModelMixin, Base):
    """
        Column     |   Type   |                         Modifiers
    ---------------+----------+-----------------------------------------------------------
     clientid      | integer  | not null default nextval('client_clientid_seq'::regclass)
     name          | text     | not null
     uname         | text     | not null
     autoprune     | smallint | default 0
     fileretention | bigint   | default 0
     jobretention  | bigint   | default 0

    Indexes:
        "client_pkey" PRIMARY KEY, btree (clientid)
        "client_name_idx" UNIQUE, btree (name)

    """

    @classmethod
    def get_list(cls, **kw):
        # SELECT client.clientid, job_bytes, max_job FROM client
        # LEFT JOIN (SELECT job.clientid, SUM(job.jobbytes) AS job_bytes FROM job
        # GROUP BY job.clientid) AS vsota ON vsota.clientid = client.clientid
        # LEFT JOIN (SELECT job.clientid, MAX(job.schedtime) AS max_job FROM job
        # GROUP BY job.clientid) AS last_job ON last_job.clientid = client.clientid;
        sum_stmt = Job.query\
            .with_entities(Job.clientid, func.sum(Job.jobbytes).label('job_sumvolbytes'))\
            .group_by(Job.clientid)\
            .subquery('stmt_sub')
        last_stmt = Job.query\
            .with_entities(Job.clientid, func.max(Job.starttime).label('job_maxschedtime')).filter(Job.jobstatus == 'T')\
            .group_by(Job.clientid)\
            .subquery('stmt_max')
        objects = cls.query.with_entities(Client, 'job_sumvolbytes', 'job_maxschedtime', func.count(Job.jobid).label('num_jobs'))\
            .outerjoin(Job, Client.clientid == Job.clientid)\
            .outerjoin(sum_stmt, sum_stmt.c.clientid == Client.clientid)\
            .outerjoin(last_stmt, last_stmt.c.clientid == Client.clientid)\
            .group_by(cls, 'job_sumvolbytes', 'job_maxschedtime')\
            .all()

        # ugly hack since sqlite returns strings for job_maxschedtime
        # TODO: report upstream to sqlalchemy
        if DBSession.bind.dialect.name == 'sqlite':
            def convert_datetime(l):
                if l.job_maxschedtime:
                    l.job_maxschedtime = datetime.datetime.strptime(l.job_maxschedtime, '%Y-%m-%d %H:%M:%S')
                return l
            objects = map(convert_datetime, objects)
        return objects

    def url(self, request):
        return request.route_url('client_detail', id=self.clientid)

    def render_name(self, request):
        return {'text': self.name, 'href': self.url(request)}

    def render_jobretention(self, request):
        # jobretention is integer of seconds
        return {'text': distance_of_time_in_words(self.jobretention)}

    def render_fileretention(self, request):
        # fileretention is integer of seconds
        return {'text': distance_of_time_in_words(self.fileretention)}

    def render_autoprune(self, request):
        return {'text': yesno(self.autoprune)}


class Job(ModelMixin, Base):
    """
         Column      |            Type             |                      Modifiers
    -----------------+-----------------------------+-----------------------------------------------------
     jobid           | integer                     | not null default nextval('job_jobid_seq'::regclass)
     job             | text                        | not null
     name            | text                        | not null
     type            | character(1)                | not null
     level           | character(1)                | not null
     clientid        | integer                     | default 0
     jobstatus       | character(1)                | not null
     schedtime       | timestamp without time zone |
     starttime       | timestamp without time zone |
     endtime         | timestamp without time zone |
     realendtime     | timestamp without time zone |
     jobtdate        | bigint                      | default 0
     volsessionid    | integer                     | default 0
     volsessiontime  | integer                     | default 0
     jobfiles        | integer                     | default 0
     jobbytes        | bigint                      | default 0
     readbytes       | bigint                      | default 0
     joberrors       | integer                     | default 0
     jobmissingfiles | integer                     | default 0
     poolid          | integer                     | default 0
     filesetid       | integer                     | default 0
     priorjobid      | integer                     | default 0
     purgedfiles     | smallint                    | default 0
     hasbase         | smallint                    | default 0
     hascache        | smallint                    | default 0
     reviewed        | smallint                    | default 0
     comment         | text                        |

    Indexes:
        "job_pkey" PRIMARY KEY, btree (jobid)
        "job_name_idx" btree (name)
    """

    schedtime = Column('schedtime', BaculaDateTime())
    starttime = Column('starttime', BaculaDateTime())
    endtime = Column('endtime', BaculaDateTime())
    realendtime = Column('realendtime', BaculaDateTime())

    status = relationship(
        "Status",
        primaryjoin="Job.jobstatus==Status.jobstatus",
        foreign_keys="Job.jobstatus",
    )
    client = relationship(
        "Client",
        primaryjoin="Job.clientid==Client.clientid",
        foreign_keys="Job.clientid",
        backref='jobs',
    )
    pool = relationship(
        "Pool",
        primaryjoin="Job.poolid==Pool.poolid",
        foreign_keys="Job.poolid",
    )
    jobmedias = relationship(
        "JobMedia",
        primaryjoin="Job.jobid==JobMedia.jobid",
        foreign_keys="JobMedia.jobid",
        backref="jobs",
    )
    logs = relationship(
        "Log",
        primaryjoin="Job.jobid==Log.jobid",
        foreign_keys="Log.jobid",
        backref="job",
    )
    files = relationship(
        "File",
        primaryjoin="Job.jobid==File.jobid",
        foreign_keys="File.jobid",
        backref="job",
    )

    @classmethod
    def get_list(cls, **kw):
        appstruct = kw.get('appstruct', {})
        if appstruct and appstruct['state'] == 'scheduled':
            return BConsole().get_upcoming_jobs()

        query = cls.query.options(joinedload(cls.status), joinedload(cls.client))
        if appstruct:
            if appstruct['status']:
                query = query.filter(cls.jobstatus == appstruct['status'])
            if appstruct['type']:
                query = query.filter(cls.type == appstruct['type'])
        return query

    @classmethod
    def get_one(cls, id_):
        query = cls.query.options(joinedload(cls.status),
                                  joinedload(cls.client),
                                  joinedload(cls.logs),
                                  joinedload(cls.pool),
                                  joinedload(cls.medias),
                                  ).get(int(id_))
        return super(Job, cls).get_one(query=query)

    @classmethod
    def get_running(cls):
        return super(Job, cls).get_list().options(joinedload(cls.status), joinedload(cls.client))\
                                         .join('status')\
                                         .filter(Status.severity == 15)\
                                         .order_by(desc(Job.starttime))

    @classmethod
    def get_last(cls):
        return super(Job, cls).get_list().options(joinedload(cls.status), joinedload(cls.client))\
                                         .join('status')\
                                         .filter(Status.severity != 15)\
                                         .order_by(desc(Job.schedtime))\
                                         .limit(5)

    def render_level(self, request):
        return {'text': LEVELS[self.level]}

    def render_type(self, request):
        return {'text': TYPES[self.type]}

    def render_status(self, request):
        if self.status:
            return {'text': self.status.jobstatuslong, 'cssclass': self.render_status_color(request)}

    def render_status_color(self, request):
        """Color of the job depending on status"""
        if not self.status:
            return ""

        if self.status.severity < 15:
            return "ok"
        elif self.status.severity > 15:
            return "error"
        else:
            return "unknown"

    @classmethod
    def get_upcoming(cls):
        return BConsole().get_upcoming_jobs()

    def url(self, request):
        return request.route_url('job_detail', id=self.jobid)

    def render_name(self, request):
        return {'text': self.name, 'href': self.url(request)}

    def render_client_name(self, request):
        if self.client:
            return self.client.render_name(request)

    def render_volume_name(self, request):
        return self.medias

    def render_pool_name(self, request):
        if self.pool:
            return self.pool.render_name(request)

    def render_duration(self, request):
        if self.starttime and self.endtime:
            return self.render_distance_of_time_in_words(self.starttime, self.endtime)

    def render_jobbytes(self, request):
        return {'text': self.format_byte_size(self.jobbytes), 'cssclass': 'nowrap'}

    def render_joberrors(self, request):
        d = {'text': self.joberrors}
        if self.joberrors:
            d['cssclass'] = 'yellow'
        return d

    def render_starttime(self, request):
        return self.render_distance_of_time_in_words(self.starttime)

    def render_jobfiles(self, request):
        return {'text': self.jobfiles}


class JobMedia(ModelMixin, Base):
    """
       Column   |  Type   |                           Modifiers
    ------------+---------+---------------------------------------------------------------
     jobmediaid | integer | not null default nextval('jobmedia_jobmediaid_seq'::regclass)
     jobid      | integer | not null
     mediaid    | integer | not null
     firstindex | integer | default 0
     lastindex  | integer | default 0
     startfile  | integer | default 0
     endfile    | integer | default 0
     startblock | bigint  | default 0
     endblock   | bigint  | default 0
     volindex   | integer | default 0

    Indexes:
        "jobmedia_pkey" PRIMARY KEY, btree (jobmediaid)
        "job_media_job_id_media_id_idx" btree (jobid, mediaid)
    """

    medias = relationship(
        "Media",
        primaryjoin="JobMedia.mediaid==Media.mediaid",
        foreign_keys="JobMedia.mediaid",
        backref="jobmedias",
    )


class Media(ModelMixin, Base):
    """
          Column      |            Type             |                        Modifiers
    ------------------+-----------------------------+---------------------------------------------------------
     actiononpurge    | smallint                    | default 0
     comment          | text                        |
     deviceid         | integer                     | default 0
     enabled          | smallint                    | default 1
     endblock         | bigint                      | default 0
     endfile          | integer                     | default 0
     firstwritten     | timestamp without time zone |
     inchanger        | smallint                    | default 0
     initialwrite     | timestamp without time zone |
     labeldate        | timestamp without time zone |
     labeltype        | integer                     | default 0
     lastwritten      | timestamp without time zone |
     locationid       | integer                     | default 0
     maxvolbytes      | bigint                      | default 0
     maxvolfiles      | integer                     | default 0
     maxvoljobs       | integer                     | default 0
     mediaaddressing  | smallint                    | default 0
     mediaid          | integer                     | not null default nextval('media_mediaid_seq'::regclass)
     mediatypeid      | integer                     | default 0
     mediatype        | text                        | not null
     poolid           | integer                     | default 0
     recyclecount     | integer                     | default 0
     recyclepoolid    | integer                     | default 0
     recycle          | smallint                    | default 0
     scratchpoolid    | integer                     | default 0
     slot             | integer                     | default 0
     storageid        | integer                     | default 0
     volblocks        | integer                     | default 0
     volbytes         | bigint                      | default 0
     volcapacitybytes | bigint                      | default 0
     volerrors        | integer                     | default 0
     volfiles         | integer                     | default 0
     voljobs          | integer                     | default 0
     volmounts        | integer                     | default 0
     volparts         | integer                     | default 0
     volreadtime      | bigint                      | default 0
     volretention     | bigint                      | default 0
     volstatus        | text                        | not null
     volumename       | text                        | not null
     voluseduration   | bigint                      | default 0
     volwrites        | integer                     | default 0
     volwritetime     | bigint                      | default 0

    Indexes:
        "media_pkey" PRIMARY KEY, btree (mediaid)
        "media_volumename_id" UNIQUE, btree (volumename)
    Check constraints:
        "media_volstatus_check" CHECK (volstatus = ANY (ARRAY['Full'::text, 'Archive'::text, 'Append'::text, 'Recycle'::text, 'Purged'::text, 'Read-Only'::text, 'Disabled'::text, 'Error'::text, 'Busy'::text, 'Used'::text, 'Cleaning'::text, 'Scratch'::text]))

    """
    firstwritten = Column('firstwritten', BaculaDateTime())
    lastwritten = Column('lastwritten', BaculaDateTime())
    labeldate = Column('labeldate', BaculaDateTime())
    initialwrite = Column('initialwrite', BaculaDateTime())

    jobs = relationship(
        "Job",
        foreign_keys="[JobMedia.mediaid, JobMedia.jobid]",
        primaryjoin="JobMedia.mediaid==Media.mediaid",
        secondaryjoin="JobMedia.jobid==Job.jobid",
        secondary=JobMedia.__table__,
        backref="medias",
    )

    storage = relationship(
        "Storage",
        primaryjoin="Media.storageid==Storage.storageid",
        foreign_keys="Media.storageid",
        backref="medias",
    )

    pool = relationship(
        "Pool",
        primaryjoin="Media.poolid==Pool.poolid",
        foreign_keys="Media.poolid",
        backref="medias",
    )

    # we don't use mediatype table yet,
    # because it only provides readonly information
    # for mediatype

    @classmethod
    def get_list(cls, **kw):
        appstruct = kw.get('appstruct', {})
        query = super(Media, cls).get_list(**kw).options(joinedload(cls.pool), joinedload(cls.storage))

        if appstruct:
            if appstruct['status']:
                query = query.filter(cls.volstatus == appstruct['status'])
            if appstruct['storage']:
                query = query.filter(cls.storageid == int(appstruct['storage']))
            if appstruct['pool']:
                query = query.filter(cls.poolid == int(appstruct['pool']))
        return query

    @classmethod
    def get_one(cls, id_):
        query = cls.query.options(joinedload(cls.pool),
                                  joinedload(cls.storage),
                                  joinedload_all('jobs.status'),
                                  joinedload_all('jobs.client'),
                                  ).get(int(id_))
        return super(Media, cls).get_one(query=query)

    def url(self, request):
        return request.route_url('volume_detail', id=self.mediaid)

    def render_volumename(self, request):
        return {'text': self.volumename, 'href': self.url(request)}

    def render_volcapacitybytes(self, request):
        return {'text': self.format_byte_size(self.volcapacitybytes), 'cssclass': 'nowrap'}

    def render_volbytes(self, request):
        return {'text': self.format_byte_size(self.volbytes), 'cssclass': 'nowrap'}

    def render_maxvolbytes(self, request):
        return {'text': self.format_byte_size(self.maxvolbytes), 'cssclass': 'nowrap'}

    def render_volretention(self, request):
        # volretention is integer of seconds
        return {'text': distance_of_time_in_words(self.volretention)}

    def render_volstatus(self, request):
        return {'text': self.volstatus, 'cssclass': VOLUME_STATUS_SEVERITY[self.volstatus]}

    def render_storage_name(self, request):
        if self.storage:
            return self.storage.render_name(request)

    def render_pool_name(self, request):
        if self.pool:
            return self.pool.render_name(request)

    def render_enabled(self, request):
        return {'text': yesno(self.enabled)}

    def render_recycle(self, request):
        return {'text': yesno(self.recycle)}

    def render_expires(self, request):
        if self.firstwritten and self.volretention:
            return {'text': self.firstwritten + datetime.timedelta(seconds=self.volretention)}


class Storage(ModelMixin, Base):
    """
       Column    |  Type   |                          Modifiers
    -------------+---------+-------------------------------------------------------------
     storageid   | integer | not null default nextval('storage_storageid_seq'::regclass)
     name        | text    | not null
     autochanger | integer | default 0

    Indexes:
        "storage_pkey" PRIMARY KEY, btree (storageid)
    """

    @classmethod
    def get_one(cls, id_):
        query = cls.query.options(joinedload_all('medias.pool')).get(int(id_))
        return super(Storage, cls).get_one(query=query)

    @classmethod
    def get_list(cls, **kw):
        return cls.query.with_entities(Storage,
                                       func.count(Media.mediaid).label('num_volumes'),
                                       #func.count(Job.jobid).label('num_jobs'),
                                       #func.count(Client.clientid).label('num_clients'),
                                       func.sum(Media.volbytes).label('total_backup_size'))\
                                .outerjoin('medias')\
                                .group_by(cls)
                                #.outerjoin('medias', 'jobmedias', 'jobs')\

    @classmethod
    def get_values(cls):
        return cls.query.with_entities(cls.storageid, cls.name).all()

    def url(self, request):
        return request.route_url('storage_detail', id=self.storageid)

    def render_name(self, request):
        return {'text': self.name, 'href': self.url(request)}

    def render_autochanger(self, request):
        return {'text': yesno(self.autochanger)}


class Log(ModelMixin, Base):
    """
     Column  |            Type             |                      Modifiers
    ---------+-----------------------------+-----------------------------------------------------
     logid   | integer                     | not null default nextval('log_logid_seq'::regclass)
     jobid   | integer                     | not null
     time    | timestamp without time zone |
     logtext | text                        | not null

    Indexes:
        "log_pkey" PRIMARY KEY, btree (logid)
        "log_name_idx" btree (jobid)

    """
    time = Column('time', BaculaDateTime())

    @classmethod
    def get_list(cls, **kw):
        appstruct = kw.get('appstruct', {})
        query = super(Log, cls).get_list(**kw).options(joinedload('job'))

        if appstruct:
            if appstruct['from_time']:
                query = query.filter(cls.time >= appstruct['from_time'])

            if appstruct['to_time']:
                query = query.filter(cls.time <= appstruct['to_time'])

        return query

    def render_jobid(self, request):
        return {'text': self.jobid, 'href': self.job.url(request)}

    def render_time(self, request):
        return {'text': self.time}

    def render_logtext(self, request):
        d = {}
        # sanitize logs
        log = self.logtext.strip('\\').strip().strip('\n')
        d['text'] = Markup(nl2br(log))
        if 'ERR' in self.logtext or 'Fatal error' in self.logtext or 'Backup Error' in self.logtext:
            d['cssclass'] = 'error'
        return d


class Pool(ModelMixin, Base):
    """
           Column       |   Type   |                       Modifiers
    --------------------+----------+-------------------------------------------------------
     acceptanyvolume    | smallint | default 0
     actiononpurge      | smallint | default 0
     autoprune          | smallint | default 0
     enabled            | smallint | default 1
     labelformat        | text     | not null
     labeltype          | integer  | default 0
     maxvolbytes        | bigint   | default 0
     maxvolfiles        | integer  | default 0
     maxvoljobs         | integer  | default 0
     maxvols            | integer  | default 0
     migrationhighbytes | bigint   | default 0
     migrationlowbytes  | bigint   | default 0
     migrationtime      | bigint   | default 0
     name               | text     | not null
     nextpoolid         | integer  | default 0
     numvols            | integer  | default 0
     poolid             | integer  | not null default nextval('pool_poolid_seq'::regclass)
     pooltype           | text     |
     recyclepoolid      | integer  | default 0
     recycle            | smallint | default 0
     scratchpoolid      | integer  | default 0
     usecatalog         | smallint | default 0
     useonce            | smallint | default 0
     volretention       | bigint   | default 0
     voluseduration     | bigint   | default 0

    Indexes:
        "pool_pkey" PRIMARY KEY, btree (poolid)
        "pool_name_idx" btree (name)
    Check constraints:
        "pool_pooltype_check" CHECK (pooltype = ANY (ARRAY['Backup'::text, 'Copy'::text, 'Cloned'::text, 'Archive'::text, 'Migration'::text, 'Scratch'::text]))
    """

    @classmethod
    def get_one(cls, id_):
        query = cls.query.options(joinedload_all('medias.storage')).get(int(id_))
        return super(Pool, cls).get_one(query=query)

    @classmethod
    def get_values(cls):
        return cls.query.with_entities(cls.poolid, cls.name).all()

    def url(self, request):
        return request.route_url('pool_detail', id=self.poolid)

    def render_name(self, request):
        return {'text': self.name, 'href': self.url(request)}

    def render_enabled(self, request):
        return {'text': yesno(self.enabled)}

    def render_usecatalog(self, request):
        return {'text': yesno(self.usecatalog)}

    def render_useonce(self, request):
        return {'text': yesno(self.useonce)}

    def render_recycle(self, request):
        return {'text': yesno(self.recycle)}

    def render_autoprune(self, request):
        return {'text': yesno(self.autoprune)}

    def render_acceptanyvolume(self, request):
        return {'text': yesno(self.acceptanyvolume)}


class FileSet(ModelMixin, Base):
    """
       Column   |            Type             |                          Modifiers
    ------------+-----------------------------+-------------------------------------------------------------
     filesetid  | integer                     | not null default nextval('fileset_filesetid_seq'::regclass)
     fileset    | text                        | not null
     md5        | text                        | not null
     createtime | timestamp without time zone | not null

    Indexes:
        "fileset_pkey" PRIMARY KEY, btree (filesetid)
        "fileset_name_idx" btree (fileset)
    """
    createtime = Column('createtime', BaculaDateTime())


class Filename(ModelMixin, Base):
    """
       Column   |  Type   |                           Modifiers
    ------------+---------+---------------------------------------------------------------
     filenameid | integer | not null default nextval('filename_filenameid_seq'::regclass)
     name       | text    | not null

    Indexes:
        "filename_pkey" PRIMARY KEY, btree (filenameid)
        "filename_name_idx" btree (name)
    """


class Path(ModelMixin, Base):
    """
     Column |  Type   |                       Modifiers
    --------+---------+-------------------------------------------------------
     pathid | integer | not null default nextval('path_pathid_seq'::regclass)
     path   | text    | not null

    Indexes:
        "path_pkey" PRIMARY KEY, btree (pathid)
        "path_name_idx" btree (path)
    """


class File(ModelMixin, Base):
    """
       Column   |  Type   |                       Modifiers
    ------------+---------+-------------------------------------------------------
     fileid     | bigint  | not null default nextval('file_fileid_seq'::regclass)
     fileindex  | integer | not null default 0
     jobid      | integer | not null
     pathid     | integer | not null
     filenameid | integer | not null
     markid     | integer | not null default 0
     lstat      | text    | not null
     md5        | text    | not null

    Indexes:
        "file_pkey" PRIMARY KEY, btree (fileid)
        "file_jobid_idx" btree (jobid)
        "file_jpfid_idx" btree (jobid, pathid, filenameid)
    """

    filename = relationship(
        "Filename",
        primaryjoin="File.filenameid==Filename.filenameid",
        foreign_keys="File.filenameid",
        innerjoin=True,
    )
    path = relationship(
        "Path",
        primaryjoin="File.pathid==Path.pathid",
        foreign_keys="File.pathid",
        innerjoin=True,
    )

    # lstat: http://docs.python.org/library/os.html?highlight=lstat#os.stat
    # dev    ino      mode   nlink  uid gid rdev bytes  blksize blocks atime       mtime       ctime       optional
    # [2049, 6160392, 16877, 2,     0,  0,  0,   16384, 4096,   32,    1304426698, 1330102044, 1330102044, 0, 0, 2]

    def get_stat_data(self):
        data = getattr(self, 'lstat_raw', None)
        if data is None:
            self.lstat_raw = [decode_base64(s) for s in self.lstat.split()]
            return self.lstat_raw
        else:
            return data

    def render_filename(self, request):
        return {'text': self.path.path + self.filename.name}

    def render_size(self, request):
        return {'text': self.format_byte_size(self.get_stat_data()[7]), 'cssclass': 'nowrap'}

    def render_uid(self, request):
        return {'text': self.get_stat_data()[4]}

    def render_gid(self, request):
        return {'text': self.get_stat_data()[5]}

    def render_mode(self, request):
        return {'text': oct(stat.S_IMODE(self.get_stat_data()[2]))}

    # TODO: pathvisibility: size and number of files in directory
    # TODO: pathhierarchy: parent-child relationship with paths



import colander
from deform_bootstrap.widget import DateTimeInputWidget, ChosenSingleWidget

from almir.models import TYPES, VOLUME_STATUS_SEVERITY

job_states = (
    ('finished+running', 'Finished + Running'),
    ('scheduled', 'Scheduled'),
)


def deferred_widget_factory(values_name):
    @colander.deferred
    def deferred_widget(node, kw):
        widget = kw.get('widget', ChosenSingleWidget)
        values = kw.get(values_name, [])
        return widget(values=values)
    return deferred_widget


class JobSchema(colander.MappingSchema):
    state = colander.SchemaNode(colander.String(),
                                name='state',
                                widget=ChosenSingleWidget(values=job_states),
                                missing=colander.null,
                                default="finished+running")
    type = colander.SchemaNode(colander.String(),
                               widget=ChosenSingleWidget(values=[('', '---')] + TYPES.items()),
                               missing=colander.null)
    status = colander.SchemaNode(colander.String(),
                                missing=colander.null,
                                 widget=deferred_widget_factory('status_values'))


class MediaSchema(colander.MappingSchema):
    status = colander.SchemaNode(colander.String(),
                                 widget=ChosenSingleWidget(values=[('', '---')] + zip(VOLUME_STATUS_SEVERITY.keys(), VOLUME_STATUS_SEVERITY.keys())),
                                 missing=colander.null)
    storage = colander.SchemaNode(colander.Integer(),
                                  missing=colander.null,
                                  widget=deferred_widget_factory('storage_values'))
    pool = colander.SchemaNode(colander.Integer(),
                               missing=colander.null,
                               widget=deferred_widget_factory('pool_values'))


class LogSchema(colander.MappingSchema):
    from_time = colander.SchemaNode(colander.DateTime(),
                                    description=u"NOTE: time should be in form of HH:MM:SS",
                                    missing=colander.null,
                                    widget=DateTimeInputWidget())
    to_time = colander.SchemaNode(colander.DateTime(),
                                  description=u"NOTE: time should be in form of HH:MM:SS",
                                  missing=colander.null,
                                  widget=DateTimeInputWidget())

# TODO: upstream: fix whitespace in .form-actions div



"""Dirty script to output buildout.cfg, but it does the job.
"""
import os
import getpass
import socket
import subprocess
import readline
import time
readline  # pyflakes: we import readline and raw_input has history!

import pytz
import sqlalchemy

from almir.lib.bconsole import BConsole


ROOT = os.path.join(os.path.dirname(__file__), '..', '..')
OUTPUT = os.path.join(ROOT, 'buildout.cfg')
TEMPLATE = """
[buildout]
extends = buildout.d/production.cfg

[almir]
# define sql database connection as specified in
# http://docs.sqlalchemy.org/en/latest/core/engines.html#database-urls
#sqla_engine_url = sqlite:///%%(here)s/bacula.db
#sqla_engine_url = postgresql+pg8000://bacula:<password>@hostname/<database>
#sqla_engine_url = mysql+mysqlconnector://bacula:<password>@hostname/<database>
sqla_engine_url = %(engine)s

# your timezone from TZ* column in table https://en.wikipedia.org/wiki/List_of_tz_database_time_zones
# if not set, defaults to system timezone
#timezone = Europe/Ljubljana
timezone = %(timezone)s

# configure http daemon
host = %(host)s
port = %(port)s

# bconsole director parameters
director_name = %(director_name)s
director_port = %(director_port)s
director_address = %(director_address)s
director_password = %(director_password)s
"""


def validate_open_port(v):
    validate_int(v)

    # python2.6 will overflow port integer if it is too big
    if int(v) not in xrange(0, 65536):
        raise ValueError('Cannot listen on port %s: port must be 0-65535.' % v)

    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        s.bind(('', int(v)))
        # port is open
    except Exception, e:
        raise ValueError('Cannot listen on port %s: %s' % (v, e))
    finally:
        s.close()


def validate_timezone(v):
    try:
        pytz.timezone(v)
    except:
        raise ValueError('Invalid timezone: %s' % v)


def validate_int(v):
    try:
        int(v)
    except ValueError:
        raise ValueError('Must be an integer!')


def validate_engine(v):
    print 'Connecting to catalog database to verify configuration ...'
    try:
        engine = None
        engine = sqlalchemy.create_engine(v)
        if 'client' not in map(lambda e: e.lower(), engine.table_names()):
            raise ValueError('Connection string has wrong parameters (could not connect to catalog database)')  # PRAGMA: no cover
    except:
        if engine and engine.dialect.name == 'sqlite':  # PRAGMA: no cover
            print
            print 'WARNING: Using sqlite, database needs to be readable by %s. Fix is usually:' % getpass.getuser()
            print '$ sudo gpassword -a %s bacula' % getpass.getuser()
            print 'You will need to relogin and restart install procedure, since permission are not updated on the fly'
            print
            # TODO: display file permissions
        raise
    print 'OK!'


def ask_question(question, default=None, validator=None, func=raw_input):
    good_answer = None

    while good_answer is None:
        answer = func("--> " + question)

        if validator and answer:
            try:
                validator(answer)
            except Exception, e:
                print '    Try again: ' + str(e)
                continue

        if answer:
            good_answer = answer
        elif not answer and default is not None:
            good_answer = default
        elif not answer and default is None:
            continue

    # add newline
    print
    return good_answer


def main():
    """Entry point of this script"""
    print '\n\nConfiguring almir ... If you want to use the default value (where possible), press enter.\n'

    options = {}
    options['host'] = ask_question('Host to listen on (default: 127.0.0.1): ', default='127.0.0.1')
    options['port'] = ask_question('Port to listen on (default: 2500): ', default='2500', validator=validate_open_port)

    print 'Define SQL database connection to bacula catalog as specified in http://docs.sqlalchemy.org/en/latest/core/engines.html#database-urls'
    print 'For example:'
    print '    postgresql+pg8000://<user>:<password>@<hostname>/<database>'
    print '    mysql+mysqlconnector://<user>:<password>@<hostname>/<database>'
    print '    sqlite:////var/lib/bacula/bacula.db'
    print '    sqlite:///bacula.db'
    print
    options['engine'] = ask_question('SQL connection string: ', validator=validate_engine)

    print 'Timezone of director in format of TZ* column in table https://en.wikipedia.org/wiki/List_of_tz_database_time_zones'
    print 'For example:'
    print '    Europe/Ljubljana'
    print '    CET'
    print

    try:
        default_timezone = " (default: %s)" % pytz.timezone(time.tzname[0])
    except:
        default_timezone = ''
    options['timezone'] = ask_question('Timezone%s' % default_timezone, default='', validator=validate_timezone)

    # TODO: in future we may extract this from bconsole config file?
    print 'Almost finished, we just need bconsole parameters to connect to director!'
    print 'Normally you will find needed information in /etc/bacula/bconsole.conf'
    print

    try:
        output = subprocess.Popen(['which', 'bconsole'], stdout=subprocess.PIPE).communicate()[0].strip()
    except subprocess.CalledProcessError:
        print 'WARNING: bconsole command is not executable from current user!'
    else:
        if not os.access(output, os.X_OK):
            print 'WARNING: bconsole command is not executable from current user!'

    bconsole_running = False
    while not bconsole_running:
        options['director_name'] = ask_question('Name of director to connect to (default: localhost-dir): ', default='localhost-dir')
        options['director_address'] = ask_question('Address of director to connect to (default: localhost): ', default='localhost')
        options['director_port'] = ask_question('Port of director to connect to (default: 9101): ', default='9101', validator=validate_open_port)
        options['director_password'] = ask_question('Password of director to connect to: ', func=getpass.getpass)

        print 'Connecting director with bconsole to verify configuration ...'

        # need a way to generate bconsole config before running this
        with BConsole.from_temp_config(name=options['director_name'],
                                       address=options['director_address'],
                                       port=options['director_port'],
                                       password=options['director_password']) as bconsole:
            if bconsole.is_running():
                print 'OK!'
                bconsole_running = True
            else:
                print 'ERROR: Could not connect to director %s, verify configuration and try again!' % options['director_name']  # PRAGMA: no cover

    with open(OUTPUT, 'w') as f:
        f.write(TEMPLATE % options)

    print
    print 'Written to %s.' % os.path.realpath(OUTPUT)


if __name__ == '__main__':
    main()  # PRAGMA: no cover
    # TODO: add notes how to proceed if we get an exception



import datetime
import re

import sqlalchemy.types as types

from almir.lib.utils import convert_timezone

DATETIME_RE = re.compile("(\d+)-(\d+)-(\d+) (\d+):(\d+):(\d+)(?:\.(\d+))?")


class BaculaDateTime(types.TypeDecorator):
    '''Changes sqlite DateTime to parse 0 values as no value. Also converts to right timezone'''

    impl = types.DateTime

    def process_result_value(self, value, dialect=None):
        return convert_timezone(value)

    def result_processor(self, dialect, coltype):
        if dialect.name == 'sqlite':
            # http://docs.sqlalchemy.org/en/latest/core/types.html?highlight=typedecorator#sqlalchemy.types.TypeDecorator
            # http://docs.sqlalchemy.org/en/latest/dialects/sqlite.html#sqlalchemy.dialects.sqlite.DATETIME
            # https://bitbucket.org/zzzeek/sqlalchemy/src/3b458030a0f3/lib/sqlalchemy/dialects/sqlite/base.py
            def process(value):
                # if we have value as 0, that should be None
                if not value:
                    return None
                else:
                    try:
                        m = DATETIME_RE.match(value)
                    except TypeError:
                        raise ValueError("Couldn't parse %s string '%r' "
                                        "- value is not a string." % (datetime.datetime_.__name__, value))
                    if m is None:
                        raise ValueError("Couldn't parse %s string: "
                                        "'%s'" % (datetime.datetime.__name__, value))
                    return self.process_result_value(datetime.datetime(*map(int, m.groups(0))))
            return process
        else:
            return super(BaculaDateTime, self).result_processor(dialect, coltype)



import os
import time
from docutils.core import publish_parts

import pytz
from pyramid.threadlocal import get_current_registry
from pyramid.config import Configurator


def convert_timezone(datetime):
    """Converts datetime to timezone aware datetime.

    Retrieving timezone:

    - get `timezone` from .ini settings
    - default to system timezone

    """
    if datetime is None:
        return None

    timezone = get_current_registry().settings.get('timezone', None)
    if not timezone:
        timezone = time.tzname[0]
    return pytz.timezone(timezone).localize(datetime)


def timedelta_to_seconds(td):
    """"""
    # http://docs.python.org/library/datetime.html?highlight=total_seconds#datetime.timedelta.total_seconds
    # to keep python2.6 support
    return (td.microseconds + (td.seconds + td.days * 24 * 3600) * 10 ** 6) / 10 ** 6


def render_rst_section(filename):
    """Finds filename in documentation directory and renders it to html."""
    path = os.path.join(os.path.dirname(os.path.abspath(__file__)), '..', '..', 'docs', 'source', filename)
    with open(path) as f:
        parts = publish_parts(f.read(), writer_name="html", settings_overrides={'initial_header_level': 2, 'doctitle_xform': False})
        return parts['html_body']


def yesno(text):
    return 'Yes' if text else 'No'


def nl2br(text):
    return text.replace('\n', '<br />')


def get_jinja_macro(macro):
    """Return actual function from a jinja2 template"""
    config = Configurator(get_current_registry())
    template = config.get_jinja2_environment().get_template('templates/macros.jinja2')
    return getattr(template.module, macro)



import fcntl
import select
import os
import re
import shlex
import tempfile
from contextlib import contextmanager
from subprocess import Popen, PIPE

from almir.lib.utils import nl2br


CURRENT_DIRECTORY = os.path.dirname(os.path.abspath(__file__))
JOBS_DEF_RE = re.compile("""
                              \s+name=(?P<name>.+)\s+JobType
                              .+
                              level=(?P<level>\S*)
                              .+
                              Priority=(?P<priority>\d*)
                              .+
                              Client
                              .+
                              FileSet:\s+name=(?P<fileset>.*)\n
                              .+
                              Storage
                              .+
                              Pool
                              """, re.X)  # TODO: where/when


class BConsoleError(Exception):
    pass


class DirectorNotRunning(BConsoleError):
    pass


class BConsole(object):
    """Interface to bconsole binary"""
    # TODO: provide bconsole session through http request to avoid multiple connects in one http request which slow down performance

    def __init__(self, bconsole_command='bconsole -n -c %s', config_file=None):
        default_config_file = os.path.realpath(os.path.join(CURRENT_DIRECTORY, '..', '..', 'bconsole.conf'))
        self.config_file = config_file or default_config_file
        self.bconsole_command = bconsole_command % self.config_file

    @classmethod
    @contextmanager
    def from_temp_config(cls, name, address, port, password):
        """Constructs :class:`BConsole` object with help of passing temporary file for the session.
        """
        with tempfile.NamedTemporaryFile() as f:
            template = os.path.join(CURRENT_DIRECTORY, '..', '..', 'buildout.d', 'templates', 'bconsole.conf.in')
            with open(template) as f2:
                config = f2.read()\
                           .replace('${almir:director_name}', name)\
                           .replace('${almir:director_port}', port)\
                           .replace('${almir:director_password}', password)\
                           .replace('${almir:director_address}', address)
            f.write(config)
            f.flush()
            yield cls(config_file=f.name)

    def start_process(self):
        return Popen(shlex.split(self.bconsole_command), stdout=PIPE, stdin=PIPE, stderr=PIPE)

    def is_running(self):
        try:
            self.get_version()
            return True
        except DirectorNotRunning:
            return False

    def get_version(self):
        p = self.start_process()
        stdout, stderr = p.communicate('version\n')
        version = filter(lambda s: 'Version' in s, stdout.split('\n'))
        if version:
            return version[-1]
        else:
            raise DirectorNotRunning

    def get_jobs_settings(self):
        p = self.start_process()
        stdout, stderr = p.communicate('show job')
        jobs = []
        for job in stdout.split('Job:'):
            jobs.append(JOBS_DEF_RE.find(stdout))
        return jobs

    def make_backup(self, job, level=None, storage=None, fileset=None, client=None, priority=None, pool=None, when=None):
        # TODO: figure out how to preselect options based on job definition (bat does it)
        # TODO: support 'where' for restore
        # TODO: support datetime for when
        # TODO: support parameters as database models instances
        cmd = 'run job=%s' % job

        if level:
            cmd += " level=%s" % level
        if storage:
            cmd += " storage=%s" % storage
        if fileset:
            cmd += " fileset=%s" % fileset
        if client:
            cmd += " client=%s" % client
        if priority:
            cmd += " priority=%s" % priority
        if pool:
            cmd += " pool=%s" % pool
        if when:
            cmd += " when=%s" % when

        p = self.start_process()
        stdout, stderr = p.communicate(cmd + "\nyes\n")
        if True:
            return "jobid"
        else:
            # TODO: stderr why job failed?
            return False

    def get_upcoming_jobs(self, days=1):
        """"""
        p = self.start_process()
        stdout, stderr = p.communicate('.status dir scheduled days=%d\n' % days)

        #if stderr.strip():
        #    pass  # TODO: display flash?

        try:
            unparsed_jobs = stdout.split('===================================================================================\n')[1].split('====\n')[0]
        except IndexError:
            return []

        jobs = []
        for line in unparsed_jobs.split('\n'):
            if not line.strip():
                continue

            jobs.append({
                         'level': line[:14].strip(),
                         'type': line[14:23].strip(),
                         'priority': line[23:28].strip(),
                         'date': line[28:38].strip(),
                         'time': line[38:44].strip(),
                         'name': line[47:67].strip(),
                         'volume': line[67:].strip(),
            })

        return jobs

    def send_command_by_polling(self, command, process=None):
        """"""
        if command == 'quit':
            return process, {'commands': ['Try harder.']}

        # start bconsole session if it's not initialized
        if process is None:
            process = self.start_process()

        poll = process.poll()
        if poll is not None:
            process = None
            return process, {'error': 'Connection to director terminated with status %d. Refresh to reconnect.' % poll}

        # send bconsole command
        if command:
            process.stdin.write(command.strip().encode('utf-8') + '\n')

        # make stdout fileobject nonblockable
        fp = process.stdout.fileno()
        flags = fcntl.fcntl(fp, fcntl.F_GETFL)
        fcntl.fcntl(fp, fcntl.F_SETFL, flags | os.O_NONBLOCK)

        output = ''

        while 1:
            # wait for data or timeout
            [i, o, e] = select.select([fp], [], [], 1)
            if i:
                # we have more data
                output += process.stdout.read(1000)
            else:
                # we have a timeout
                output = nl2br(output)

                return process, {"commands": [output]}



import time
import unittest2
from subprocess import Popen, PIPE

from mock import patch

from almir.lib.bconsole import BConsole


class TestBConsole(unittest2.TestCase):

    def test_is_running(self):
        b = BConsole()
        with patch.object(b, 'start_process') as mock_method:
            start_process = mock_method.return_value
            start_process.communicate.return_value = ('Version', '')
            self.assertTrue(b.is_running())

        with patch.object(b, 'start_process') as mock_method:
            start_process = mock_method.return_value
            start_process.communicate.return_value = ('error', 'error')
            self.assertFalse(b.is_running())

    def test_from_temp_config(self):
        output = \
"""# generated by almir, you should never edit this file. Do:
# vim buildout.cfg
# bin/buildout
# bin/supervisorctl restart all

Director {
    Name = test
    DIRport = 12345
    address = add
    Password = "qweqwe"
}
"""
        with BConsole.from_temp_config(
            name='test',
            address='add',
            port='12345',
            password='qweqwe'
        ) as b:
            expected = open(b.config_file).read()
            self.assertEqual(expected.replace(' ', '').replace('\r', ''),
                             output.replace(' ', '').replace('\r', ''))

    def test_get_upcoming_jobs(self):
        b = BConsole()
        with patch.object(b, 'start_process') as mock_method:
            start_process = mock_method.return_value
            start_process.communicate.return_value = ("""
Scheduled Jobs:
Level          Type     Pri  Scheduled          Name               Volume
===================================================================================
               Admin      8  18-Apr-12 20:30    UpdateSlots
Differential   Backup    10  18-Mar-12 23:05    BackupClient1      *unknown*
Full           Backup    11  18-Mar-12 23:10    BackupCatalog      *unknown*
====
""", '')

            jobs = b.get_upcoming_jobs()
            self.assertEqual(jobs, [{'date': '18-Apr-12',
                                     'level': '',
                                     'name': 'UpdateSlots',
                                     'time': '20:30',
                                     'priority': '8',
                                     'type': 'Admin',
                                     'volume': ''},
                                    {'date': '18-Mar-12',
                                     'level': 'Differential',
                                     'name': 'BackupClient1',
                                     'priority': '10',
                                     'time': '23:05',
                                     'type': 'Backup',
                                     'volume': '*unknown*'},
                                    {'date': '18-Mar-12',
                                     'level': 'Full',
                                     'name': 'BackupCatalog',
                                     'priority': '11',
                                     'time': '23:10',
                                     'type': 'Backup',
                                     'volume': '*unknown*'}])

    def test_send_command_by_polling(self):
        b = BConsole()
        with patch.object(b, 'start_process') as mock_method:
            mock_method.return_value = Popen(['cat'], stdout=PIPE, stdin=PIPE, stderr=PIPE)

            process, outputs = b.send_command_by_polling('version')
            self.assertEqual(outputs['commands'][0], 'version<br />')

            process.kill()
            time.sleep(1)
            process, outputs = b.send_command_by_polling('version', process)
            self.assertTrue('error' in outputs)

        with patch.object(b, 'start_process') as mock_method:
            start_process = mock_method.return_value
            start_process.communicate.return_value = ('error', 'error')

            process, outputs = b.send_command_by_polling('quit')
            self.assertEqual(outputs['commands'][0], 'Try harder.')



import os

from setuptools import setup, find_packages


here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.rst')).read()
CHANGES = open(os.path.join(here, 'docs', 'source', 'changelog.rst')).read()

requires = [
    'pyramid',
    'pyramid_tm',
    'pyramid_jinja2',
    'pyramid_exclog',
    'pyramid_beaker',
    'colander',
    'deform',
    'deform_bootstrap',
    'WebHelpers',
    'transaction',
    'SQLAlchemy',
    'zope.sqlalchemy',
    'waitress',
    'docutils',
    'pg8000',  # postgres
    'mysql-connector-repackaged',  # original one does not install with buildout
    'pytz',
]

try:
    import sqlite3
    sqlite3  # pyflakes
except ImportError:
    requires.append('pysqlite')

setup(name='almir',
      version='0.1.3',
      description='almir',
      long_description=README + '\n\n' + CHANGES,
      classifiers=[
          "Programming Language :: Python",
          "Framework :: Pylons",
          "Topic :: Internet :: WWW/HTTP",
          "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
          "Topic :: System :: Archiving :: Backup",
          "License :: OSI Approved :: GNU General Public License (GPL)",
      ],
      author=u'Domen Kozar',
      author_email='domen@dev.si',
      url='',
      keywords='web pyramid bacula administration',
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False,
      test_suite='almir',
      install_requires=requires,
      entry_points="""
      [paste.app_factory]
      main = almir:main

      [console_scripts]
      almir_parse_console_commands = almir.scripts.parse_console_commands:main
      """,
      paster_plugins=['pyramid'],
      extras_require={
          'test': [
              'webtest',
              'nose',
              'coverage',
              'unittest2',
              'mock',
              'tissue',
          ],
          'develop': [
              'bpython',
              'z3c.checkversions [buildout]',
              'jarn.mkrelease',
              'waitress',
              'pyramid_debugtoolbar',
              'Sphinx',
              'docutils',
          ],
      },
)



import re
import os
import argparse


regex = re.compile(r"""
   {\sNT_\("([^\)]+)"\)
   .+
   _\("([^)]+)"\)
   [,\s]+
   NT_\("([^)]+)"\)
""", re.X | re.M)


def parse_console_commands(source):
    with open(source, 'r') as f:
        text = f.read()
        match = regex.findall(text)

    __here__ = os.path.dirname(os.path.abspath(__file__))
    output = os.path.join(__here__, '..', 'lib', 'console_commands.py')

    with open(output, 'w') as f:
        f.write('# generated by almir/scripts/parse_console_commands.py\n')
        f.write('CONSOLE_COMMANDS = {\n')
        for name, desc, help_ in match:
            help_ = help_.replace('|', 'OR')
            f.write('    "%s": {"desc": "%s", "help": "%s"},\n' % (name, desc, help_))
        f.write('}')
        print 'written %s' % output


def main():
    parser = argparse.ArgumentParser(prog='parse_console_commands')
    parser.add_argument('bacula_source_directory')
    args = parser.parse_args()
    parse_console_commands(os.path.join(args.bacula_source_directory, 'bacula/src/dird/ua_cmds.c'))



# generated by almir/scripts/parse_console_commands.py
CONSOLE_COMMANDS = {
    "add": {"desc": "Add media to a pool", "help": "pool=<pool-name> storage=<storage> jobid=<JobId>"},
    "autodisplay": {"desc": "Autodisplay console messages", "help": "on OR off"},
    "automount": {"desc": "Automount after label", "help": "on OR off"},
    "cancel": {"desc": "Cancel a job", "help": "jobid=<number> job=<job-name> ujobid=<unique-jobid>"},
    "create": {"desc": "Create DB Pool from resource", "help": "pool=<pool-name>"},
    "delete": {"desc": "Delete volume, pool or job", "help": "volume=<vol-name> pool=<pool-name> jobid=<id>"},
    "disable": {"desc": "Disable a job", "help": "job=<name>"},
    "enable": {"desc": "Enable a job", "help": "job=<name>"},
    "estimate": {"desc": "Performs FileSet estimate, listing gives full listing", "help": "fileset=<fs> client=<cli> level=<level> accurate=<yes/no> job=<job> listing"},
    "gui": {"desc": "Non-interactive gui mode", "help": "on OR off"},
    "help": {"desc": "Print help on specific command", "help": "add autodisplay automount cancel create delete disable\n\tenable estimate exit gui label list llist"
       "\n\tmessages memory mount prune purge python quit query\n\trestore relabel release reload run status"
       "\n\tsetdebug setip show sqlquery time trace unmount\n\tumount update use var version wait"},
    "label": {"desc": "Label a tape", "help": "storage=<storage> volume=<vol> pool=<pool> slot=<slot> barcodes"},
    "list": {"desc": "List objects from catalog", "help": "pools OR jobs OR jobtotals OR volume OR media <pool=pool-name> OR files jobid=<nn> OR copies jobid=<nn>"},
    "llist": {"desc": "Full or long list like list command", "help": "pools OR jobs OR jobtotals OR media <pool=pool-name> OR files jobid=<nn> OR copies jobid=<nn>"},
    "mount": {"desc": "Mount storage", "help": "storage=<storage-name> slot=<num> drive=<num> [ jobid=<id> OR job=<job-name> ]"},
    "prune": {"desc": "Prune expired records from catalog", "help": "files OR jobs OR pool=<pool> OR client=<client-name> OR volume=<volume-name> "},
    "purge": {"desc": "Purge records from catalog", "help": "files jobs volume=<vol> [action=<action> devicetype=<type> pool=<pool> allpools storage=<st> drive=<num>]"},
    "restore": {"desc": "Restore files", "help": "where=</path> client=<client> storage=<storage> bootstrap=<file> "
       "restore_job=<job>"
       "\n\tcomment=<text> jobid=<jobid> done select all"},
    "relabel": {"desc": "Relabel a tape", "help": "storage=<storage-name> oldvolume=<old-volume-name>\n\tvolume=<newvolume-name> pool=<pool>"},
    "release": {"desc": "Release storage", "help": "storage=<storage-name>"},
    "run": {"desc": "Run a job", "help": "job=<job-name> client=<client-name>\n\tfileset=<FileSet-name> level=<level-keyword>\n\tstorage=<storage-name>"
       "where=<directory-prefix>\n\twhen=<universal-time-specification>\n\tcomment=<text> yes"},
    "status": {"desc": "Report status", "help": "all OR dir=<dir-name> OR director OR client=<client-name> OR storage=<storage-name> slots OR days=nnn"},
    "setdebug": {"desc": "Sets debug level", "help": "level=<nn> trace=0/1 client=<client-name> OR dir OR storage=<storage-name> OR all"},
    "show": {"desc": "Show resource records", "help": "job=<xxx> OR  pool=<yyy> OR fileset=<aaa> schedule=<sss> OR client=<zzz> OR disabled OR all"},
    "trace": {"desc": "Turn on/off trace to file", "help": "on OR off"},
    "unmount": {"desc": "Unmount storage", "help": "storage=<storage-name> [ drive=<num> ] OR jobid=<id> OR job=<job-name>"},
    "umount": {"desc": "Umount - for old-time Unix guys, see unmount", "help": "storage=<storage-name> [ drive=<num> ] OR jobid=<id> OR job=<job-name>"},
    "update": {"desc": "Update volume, pool or stats", "help": "stats\n\tpool=<poolname>\n\tslots storage=<storage> scan"
       "\n\tvolume=<volname> volstatus=<status> volretention=<time-def>"
       "\n\t pool=<pool> recycle=<yes/no> slot=<number>\n\t inchanger=<yes/no>"
       "\n\t maxvolbytes=<size> maxvolfiles=<nb> maxvoljobs=<nb>"
       "\n\t enable=<yes/no> recyclepool=<pool> actiononpurge=<action>"},
    "use": {"desc": "Use catalog xxx", "help": "catalog=<catalog>"},
    "wait": {"desc": "Wait until no jobs are running", "help": "jobname=<name> OR jobid=<nnn> OR ujobid=<complete_name>"},
}
